@@ -2,16 +2,18 @@ from typing import Union, Dict, Iterable, Tuple, cast, List, Optional, Generator
 from collections import defaultdict
 from enum import Enum
 import warnings
+import bisect
 
 import numpy as np
+import sympy.ntheory
 
-
-from qupulse._program.waveforms import Waveform
+from qupulse._program.waveforms import Waveform, ConstantWaveform
 from qupulse._program.volatile import VolatileRepetitionCount, VolatileProperty
 
 from qupulse.utils import is_integer
 from qupulse.utils.types import TimeType, MeasurementWindow
 from qupulse.utils.tree import Node, is_tree_circular
+from qupulse.utils.numeric import smallest_factor_ge
 
 from qupulse._program.waveforms import SequenceWaveform, RepetitionWaveform
 
@@ -229,9 +231,13 @@ class Loop(Node):
                           measurements=None if self._measurements is None else list(self._measurements),
                           children=(child.copy_tree_structure() for child in self))
 
-    def _get_measurement_windows(self) -> Mapping[str, np.ndarray]:
+    def _get_measurement_windows(self, drop: bool) -> Mapping[str, np.ndarray]:
         """Private implementation of get_measurement_windows with a slightly different data format for easier tiling.
 
+        Args:
+            drop: Drops the measurements from the Loop i.e. the Loop will no longer have measurements attached after
+            collecting them
+
         Returns:
              A dictionary (measurement_name -> array) with begin == array[:, 0] and length == array[:, 1]
         """
@@ -243,46 +249,43 @@ class Loop(Node):
             for mw_name, begin_length_list in temp_meas_windows.items():
                 temp_meas_windows[mw_name] = [np.asarray(begin_length_list, dtype=float)]
 
+            if drop:
+                self._measurements = None
+
         # calculate duration together with meas windows in the same iteration
         if self.is_leaf():
             body_duration = float(self.body_duration)
         else:
             offset = TimeType(0)
             for child in self:
-                for mw_name, begins_length_array in child._get_measurement_windows().items():
+                for mw_name, begins_length_array in child._get_measurement_windows(drop).items():
                     begins_length_array[:, 0] += float(offset)
                     temp_meas_windows[mw_name].append(begins_length_array)
                 offset += child.duration
 
             body_duration = float(offset)
 
-        # this gives us regular dict behaviour of the returned object
-        temp_meas_windows.default_factory = None
+        # formatting like this for easier debugging
+        result = {}
 
         # repeat and add repetition based offset
         for mw_name, begin_length_list in temp_meas_windows.items():
-            temp_begin_length_array = np.concatenate(begin_length_list)
+            result[mw_name] = _repeat_loop_measurements(begin_length_list, self.repetition_count, body_duration)
 
-            begin_length_array = np.tile(temp_begin_length_array, (self.repetition_count, 1))
+        return result
 
-            shaped_begin_length_array = np.reshape(begin_length_array, (self.repetition_count, -1, 2))
-
-            shaped_begin_length_array[:, :, 0] += (np.arange(self.repetition_count) * body_duration)[:, np.newaxis]
-
-            temp_meas_windows[mw_name] = begin_length_array
-
-        # the cast is here because static type analysis struggles to detect that we replace _all_ values by ndarray in
-        # the previous loop
-        return cast(Mapping[str, np.ndarray], temp_meas_windows)
-
-    def get_measurement_windows(self) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
+    def get_measurement_windows(self, drop=False) -> Dict[str, Tuple[np.ndarray, np.ndarray]]:
         """Iterates over all children and collect the begin and length arrays of each measurement window.
 
+        Args:
+            drop: Drops the measurements from the Loop i.e. the Loop will no longer have measurements attached after
+            collecting them.
+
         Returns:
             A dictionary (measurement_name -> (begin, length)) with begin and length being :class:`numpy.ndarray`
         """
         return {mw_name: (begin_length_list[:, 0], begin_length_list[:, 1])
-                for mw_name, begin_length_list in self._get_measurement_windows().items()}
+                for mw_name, begin_length_list in self._get_measurement_windows(drop=drop).items()}
 
     def split_one_child(self, child_index=None) -> None:
         """Take the last child that has a repetition count larger one, decrease it's repetition count and insert a copy
@@ -448,6 +451,20 @@ class Loop(Node):
         else:
             return self.repetition_count, tuple(child.get_duration_structure() for child in self)
 
+    def reverse_inplace(self):
+        if self.is_leaf():
+            self._waveform = self._waveform.reversed()
+        else:
+            self._reverse_children()
+            for child in self:
+                child.reverse_inplace()
+        if self._measurements:
+            duration = self.duration
+            self._measurements = [
+                (name, duration - (begin + length), length)
+                for name, begin, length in self._measurements
+            ]
+
 
 class ChannelSplit(Exception):
     def __init__(self, channel_sets):
@@ -459,15 +476,16 @@ def to_waveform(program: Loop) -> Waveform:
         if program.repetition_count == 1:
             return program.waveform
         else:
-            return RepetitionWaveform(program.waveform, program.repetition_count)
+            return RepetitionWaveform.from_repetition_count(program.waveform, program.repetition_count)
     else:
         if len(program) == 1:
             sequenced_waveform = to_waveform(cast(Loop, program[0]))
         else:
-            sequenced_waveform = SequenceWaveform([to_waveform(cast(Loop, sub_program))
-                                                   for sub_program in program])
+            sequenced_waveform = SequenceWaveform.from_sequence(
+                [to_waveform(cast(Loop, sub_program))
+                 for sub_program in program])
         if program.repetition_count > 1:
-            return RepetitionWaveform(sequenced_waveform, program.repetition_count)
+            return RepetitionWaveform.from_repetition_count(sequenced_waveform, program.repetition_count)
         else:
             return sequenced_waveform
 
@@ -578,6 +596,82 @@ def make_compatible(program: Loop, minimal_waveform_length: int, waveform_quantu
         assert comp_level == _CompatibilityLevel.compatible
 
 
+def roll_constant_waveforms(program: Loop, minimal_waveform_quanta: int, waveform_quantum: int, sample_rate: TimeType):
+    """This function finds waveforms in program that can be replaced with repetitions of shorter waveforms and replaces
+    them. Complexity O(N_waveforms). Drops measurements because they are not correctly handled here for performance
+    reasons.
+
+    This is possible if:
+     - The waveform is constant on all channels
+     - waveform.duration * sample_rate / waveform_quantum has a factor that is bigger than minimal_waveform_quanta
+
+    Args:
+        program:
+        minimal_waveform_quanta:
+        waveform_quantum:
+        sample_rate:
+
+    Warnings:
+        DroppedMeasurementWarning: This warning is raised if a measurement is dropped.
+    """
+    if program._measurements:
+        warnings.warn("Dropping measurements. Remove measurements before calling roll_constant_waveforms by calling"
+                      " get_measurement_windows(drop=True).", category=DroppedMeasurementWarning)
+        program._measurements = None
+
+    waveform = program.waveform
+
+    if waveform is None:
+        for child in program:
+            roll_constant_waveforms(child, minimal_waveform_quanta, waveform_quantum, sample_rate)
+    else:
+        waveform_quanta = (waveform.duration * sample_rate) // waveform_quantum
+
+        # example
+        # waveform_quanta = 15
+        # minimal_waveform_quanta = 2
+        # => repetition_count = 5, new_waveform_quanta = 3
+        if waveform_quanta < minimal_waveform_quanta * 2:
+            # there is no way to roll this waveform because it is too short
+            return
+
+        const_values = waveform.constant_value_dict()
+        if const_values is None:
+            # The waveform is not constant
+            return
+
+        new_waveform_quanta = smallest_factor_ge(waveform_quanta, min_factor=minimal_waveform_quanta)
+        if new_waveform_quanta == waveform_quanta:
+            # the waveform duration in samples has no suitable factor
+            # TODO: Option to insert multiple Loop objects
+            return
+
+        additional_repetition_count = waveform_quanta // new_waveform_quanta
+
+        new_waveform = ConstantWaveform.from_mapping(
+            duration=waveform_quantum * new_waveform_quanta / sample_rate,
+            constant_values=const_values)
+
+        # use the private properties to avoid invalidating the duration cache of the parent loop
+        program._repetition_definition = program.repetition_definition * additional_repetition_count
+        program._waveform = new_waveform
+
+
+def _repeat_loop_measurements(begin_length_list: List[np.ndarray],
+                              repetition_count: int,
+                              body_duration: float
+                              ) -> np.ndarray:
+    temp_begin_length_array = np.concatenate(begin_length_list)
+
+    begin_length_array = np.tile(temp_begin_length_array, (repetition_count, 1))
+
+    shaped_begin_length_array = np.reshape(begin_length_array, (repetition_count, -1, 2))
+
+    shaped_begin_length_array[:, :, 0] += (np.arange(repetition_count) * body_duration)[:, np.newaxis]
+
+    return begin_length_array
+
+
 class MakeCompatibleWarning(ResourceWarning):
     pass
 
