@@ -11,7 +11,6 @@ import numpy as np
 from qupulse.utils.types import ChannelID, TimeType
 from qupulse.hardware.awgs.base import ProgramEntry
 from qupulse.hardware.util import get_sample_times, voltage_to_uint16
-from qupulse.pulses.parameters import Parameter
 from qupulse._program.waveforms import Waveform
 from qupulse._program._loop import Loop
 from qupulse._program.volatile import VolatileRepetitionCount, VolatileProperty
@@ -389,7 +388,8 @@ class TaborProgram(ProgramEntry):
                  offsets: Tuple[float, float],
                  voltage_transformations: Tuple[Optional[callable], Optional[callable]],
                  sample_rate: TimeType,
-                 mode: TaborSequencing = None
+                 mode: TaborSequencing = None,
+                 repetition_mode: str = "infinite",
                  ):
         if len(channels) != device_properties['chan_per_part']:
             raise TaborException('TaborProgram only supports {} channels'.format(device_properties['chan_per_part']))
@@ -420,6 +420,7 @@ class TaborProgram(ProgramEntry):
         self._parsed_program = None  # type: Optional[ParsedProgram]
         self._mode = None
         self._device_properties = device_properties
+        self._repetition_mode = repetition_mode
 
         assert mode in (TaborSequencing.ADVANCED, TaborSequencing.SINGLE), "Invalid mode"
         if mode == TaborSequencing.SINGLE:
