@@ -2,6 +2,7 @@
 This module defines the class Expression to represent mathematical expression as well as
 corresponding exception classes.
 """
+import operator
 from typing import Any, Dict, Union, Sequence, Callable, TypeVar, Type, Mapping
 from numbers import Number
 import warnings
@@ -14,7 +15,7 @@ import numpy
 
 from qupulse.serialization import AnonymousSerializable
 from qupulse.utils.sympy import sympify, to_numpy, recursive_substitution, evaluate_lambdified,\
-    get_most_simple_representation, get_variables
+    get_most_simple_representation, get_variables, evaluate_lamdified_exact_rational
 from qupulse.utils.types import TimeType
 
 __all__ = ["Expression", "ExpressionVariableMissingException", "ExpressionScalar", "ExpressionVector", "ExpressionLike"]
@@ -23,6 +24,58 @@ __all__ = ["Expression", "ExpressionVariableMissingException", "ExpressionScalar
 _ExpressionType = TypeVar('_ExpressionType', bound='Expression')
 
 
+ALLOWED_NUMERIC_SCALAR_TYPES = (float, numpy.number, int, complex, bool, numpy.bool_, TimeType)
+
+
+def _parse_evaluate_numeric(result) -> Union[Number, numpy.ndarray]:
+    """Tries to parse the result as a scalar if possible. Falls back to an array otherwise.
+    Raises:
+        ValueError if scalar result is not parsable
+    """
+    allowed_scalar = ALLOWED_NUMERIC_SCALAR_TYPES
+
+    if isinstance(result, allowed_scalar):
+        # fast path for regular evaluations
+        return result
+    if isinstance(result, tuple):
+        result, = result
+    elif isinstance(result, numpy.ndarray):
+        result = result[()]
+
+    if isinstance(result, allowed_scalar):
+        return result
+    if isinstance(result, sympy.Float):
+        return float(result)
+    elif isinstance(result, sympy.Integer):
+        return int(result)
+
+    if isinstance(result, numpy.ndarray):
+        # allow numeric vector values
+        return _parse_evaluate_numeric_vector(result)
+    raise ValueError("Non numeric result", result)
+
+
+def _parse_evaluate_numeric_vector(vector_result: numpy.ndarray) -> numpy.ndarray:
+    allowed_scalar = ALLOWED_NUMERIC_SCALAR_TYPES
+    if not issubclass(vector_result.dtype.type, allowed_scalar):
+        obj_types = set(map(type, vector_result.flat))
+        if all(issubclass(obj_type, sympy.Integer) for obj_type in obj_types):
+            result = vector_result.astype(numpy.int64)
+        elif all(issubclass(obj_type, (sympy.Integer, sympy.Float)) for obj_type in obj_types):
+            result = vector_result.astype(float)
+        else:
+            raise ValueError("Could not parse vector result", vector_result)
+    return vector_result
+
+
+def _flat_iter(arr):
+    if len(arr.shape) > 1:
+        for sub_arr in arr:
+            yield from _flat_iter(sub_arr)
+    else:
+        yield from arr
+
+
 class _ExpressionMeta(type):
     """Metaclass that forwards calls to Expression(...) to Expression.make(...) to make subclass objects"""
     def __call__(cls: Type[_ExpressionType], *args, **kwargs) -> _ExpressionType:
@@ -47,32 +100,6 @@ class Expression(AnonymousSerializable, metaclass=_ExpressionMeta):
             else:
                 raise ExpressionVariableMissingException(key_error.args[0], self) from key_error
 
-    def _parse_evaluate_numeric_result(self,
-                                       result: Union[Number, numpy.ndarray],
-                                       call_arguments: Any) -> Union[Number, numpy.ndarray]:
-        allowed_types = (float, numpy.number, int, complex, bool, numpy.bool_, TimeType)
-        if isinstance(result, tuple):
-            result = numpy.array(result)
-        if isinstance(result, numpy.ndarray):
-            if issubclass(result.dtype.type, allowed_types):
-                return result
-            else:
-                obj_types = set(map(type, result.flat))
-                if obj_types == {sympy.Float} or obj_types == {sympy.Float, sympy.Integer}:
-                    return result.astype(float)
-                elif obj_types == {sympy.Integer}:
-                    return result.astype(numpy.int64)
-                else:
-                    raise NonNumericEvaluation(self, result, call_arguments)
-        elif isinstance(result, allowed_types):
-            return result
-        elif isinstance(result, sympy.Float):
-            return float(result)
-        elif isinstance(result, sympy.Integer):
-            return int(result)
-        else:
-            raise NonNumericEvaluation(self, result, call_arguments)
-
     def evaluate_in_scope(self, scope: Mapping) -> Union[Number, numpy.ndarray]:
         """Evaluate the expression by taking the variables from the given scope (typically of type Scope but it can be
         any mapping.)
@@ -82,20 +109,10 @@ class Expression(AnonymousSerializable, metaclass=_ExpressionMeta):
         Returns:
 
         """
-        parsed_kwargs = self._parse_evaluate_numeric_arguments(scope)
-
-        result, self._expression_lambda = evaluate_lambdified(self.underlying_expression, self.variables,
-                                                              parsed_kwargs, lambdified=self._expression_lambda)
-
-        return self._parse_evaluate_numeric_result(result, scope)
+        raise NotImplementedError("")
 
     def evaluate_numeric(self, **kwargs) -> Union[Number, numpy.ndarray]:
-        parsed_kwargs = self._parse_evaluate_numeric_arguments(kwargs)
-
-        result, self._expression_lambda = evaluate_lambdified(self.underlying_expression, self.variables,
-                                                              parsed_kwargs, lambdified=self._expression_lambda)
-
-        return self._parse_evaluate_numeric_result(result, kwargs)
+        return self.evaluate_in_scope(kwargs)
 
     def __float__(self):
         if self.variables:
@@ -103,10 +120,20 @@ class Expression(AnonymousSerializable, metaclass=_ExpressionMeta):
         else:
             e = self.evaluate_numeric()
             return float(e)
-    
+
     def evaluate_symbolic(self, substitutions: Mapping[Any, Any]) -> 'Expression':
+        if len(substitutions)==0:
+            return self
         return Expression.make(recursive_substitution(sympify(self.underlying_expression), substitutions))
 
+    def _evaluate_to_time_dependent(self, scope: Mapping) -> Union['Expression', Number, numpy.ndarray]:
+        try:
+            return self.evaluate_numeric(**scope, t=sympy.symbols('t'))
+        except NonNumericEvaluation as non_num:
+            return ExpressionScalar(non_num.non_numeric_result)
+        except TypeError:
+            return self.evaluate_symbolic(scope)
+
     @property
     def variables(self) -> Sequence[str]:
         """ Get all free variables in the expression.
@@ -153,75 +180,98 @@ class ExpressionVector(Expression):
 
     def __init__(self, expression_vector: Sequence):
         super().__init__()
+
         if isinstance(expression_vector, sympy.NDimArray):
-            expression_vector = to_numpy(expression_vector)
-        self._expression_vector = self.sympify_vector(expression_vector)
-        variables = set(itertools.chain.from_iterable(map(get_variables, self._expression_vector.flat)))
-        self._variables = tuple(variables)
+            expression_shape = expression_vector.shape
+            expression_items = tuple(_flat_iter(expression_vector))
+        else:
+            expression_ndarray = self.sympify_vector(expression_vector)
+            expression_items = tuple(expression_ndarray.flat)
+            expression_shape = expression_ndarray.shape
 
-    @property
-    def expression_lambda(self) -> Callable:
-        if self._expression_lambda is None:
-            expression_lambda = sympy.lambdify(self.variables, self.underlying_expression,
-                                                     [{'ceiling': ceiling}, 'numpy'])
+        self._expression_items = expression_items
+        self._expression_shape = expression_shape
 
-            @functools.wraps(expression_lambda)
-            def expression_wrapper(*args, **kwargs):
-                result = expression_lambda(*args, **kwargs)
-                if isinstance(result, sympy.NDimArray):
-                    return numpy.array(result.tolist())
-                elif isinstance(result, list):
-                    return numpy.array(result).reshape(self.underlying_expression.shape)
-                else:
-                    return result.reshape(self.underlying_expression.shape)
-            self._expression_lambda = expression_wrapper
-        return self._expression_lambda
+        self._lambdified_items = [None] * len(self._expression_items)
+
+        variables = set(itertools.chain.from_iterable(map(get_variables, self._expression_items)))
+        self._variables = tuple(sorted(variables))
 
     @property
     def variables(self) -> Sequence[str]:
         return self._variables
 
-    def evaluate_numeric(self, **kwargs) -> Union[numpy.ndarray, Number]:
-        parsed_kwargs = self._parse_evaluate_numeric_arguments(kwargs)
-
-        result, self._expression_lambda = evaluate_lambdified(self.underlying_expression, self.variables,
-                                                              parsed_kwargs, lambdified=self._expression_lambda)
-
-        if isinstance(result, (list, tuple)):
-            result = numpy.array(result)
-
-        return self._parse_evaluate_numeric_result(numpy.array(result), kwargs)
+    def evaluate_in_scope(self, scope: Mapping) -> numpy.ndarray:
+        parsed_kwargs = self._parse_evaluate_numeric_arguments(scope)
+        flat_result = []
+        for idx, expr in enumerate(self._expression_items):
+            result, self._lambdified_items[idx] = evaluate_lambdified(expr, self.variables, parsed_kwargs,
+                                                                      lambdified=self._lambdified_items[idx])
+            flat_result.append(result)
+        result = numpy.array(flat_result).reshape(self._expression_shape)
+        try:
+            return _parse_evaluate_numeric_vector(result)
+        except ValueError as err:
+            raise NonNumericEvaluation(self, result, scope) from err
 
     def get_serialization_data(self) -> Sequence[str]:
-        def nested_get_most_simple_representation(list_or_expression):
-            if isinstance(list_or_expression, list):
-                return [nested_get_most_simple_representation(entry)
-                        for entry in list_or_expression]
-            else:
-                return get_most_simple_representation(list_or_expression)
-        return nested_get_most_simple_representation(self._expression_vector.tolist())
+        serialized_items = list(map(get_most_simple_representation, self._expression_items))
+        if len(self._expression_shape) == 0:
+            return serialized_items[0]
+        elif len(self._expression_shape) == 1:
+            return serialized_items
+        else:
+            return numpy.array(serialized_items).reshape(self._expression_shape).tolist()
+
+    def __getstate__(self):
+        return self.get_serialization_data()
+
+    def __setstate__(self, state):
+        self.__init__(state)
 
     def __str__(self):
         return str(self.get_serialization_data())
 
     def __repr__(self):
-        return 'ExpressionVector({})'.format(repr(self.get_serialization_data()))
+        return f'ExpressionVector({self.get_serialization_data()!r})'
+
+    def _sympy_(self):
+        return sympy.NDimArray(self.to_ndarray())
 
     def __eq__(self, other):
         if not isinstance(other, Expression):
-            other = Expression.make(other)
+            try:
+                other = Expression.make(other)
+            except (ValueError, TypeError):
+                return NotImplemented
         if isinstance(other, ExpressionScalar):
-            return self._expression_vector.size == 1 and self._expression_vector[0] == other.underlying_expression
-        if isinstance(other, ExpressionVector) and self._expression_vector.shape != other._expression_vector.shape:
-            return False
-        return numpy.all(self._expression_vector == other.underlying_expression)
+            return self._expression_shape in ((), (1,)) and self._expression_items[0] == other.sympified_expression
+        else:
+            return self._expression_shape == other._expression_shape and \
+                   self._expression_items == other._expression_items
+
+    def __hash__(self):
+        if self._expression_shape in ((), (1,)):
+            return hash(self._expression_items[0])
+        else:
+            return hash((self._expression_items, self._expression_shape))
 
     def __getitem__(self, item) -> Expression:
-        return self._expression_vector[item]
+        if len(self._expression_shape) == 0:
+            assert item == ()
+            expr, = self._expression_items
+            return ExpressionScalar(expr)
+        if len(self._expression_shape) == 1:
+            return ExpressionScalar(self._expression_items[item])
+        else:
+            return ExpressionVector(self.to_ndarray()[item])
+
+    def to_ndarray(self) -> numpy.ndarray:
+        return numpy.array(self._expression_items).reshape(self._expression_shape)
 
     @property
     def underlying_expression(self) -> numpy.ndarray:
-        return self._expression_vector
+        return self.to_ndarray()
 
 
 class ExpressionScalar(Expression):
@@ -243,13 +293,31 @@ class ExpressionScalar(Expression):
         super().__init__()
 
         if isinstance(ex, sympy.Expr):
-            self._original_expression = str(ex)
+            self._original_expression = None
             self._sympified_expression = ex
+            self._variables = get_variables(self._sympified_expression)
+        elif isinstance(ex, ExpressionScalar):
+            self._original_expression = ex._original_expression
+            self._sympified_expression = ex._sympified_expression            
+            self._variables = ex._variables
+        elif isinstance(ex, (int, float)):
+            if isinstance(ex, numpy.float64):
+                ex = float(ex)
+            self._original_expression = ex
+            self._sympified_expression = sympify(ex)
+            self._variables = ()
         else:
             self._original_expression = ex
             self._sympified_expression = sympify(ex)
+            self._variables = get_variables(self._sympified_expression)
 
-        self._variables = get_variables(self._sympified_expression)
+        self._exact_rational_lambdified = None
+
+    def __float__(self):
+        if isinstance(self._original_expression, float):
+            return self._original_expression
+        else:
+            return super().__float__()
 
     @property
     def underlying_expression(self) -> sympy.Expr:
@@ -259,7 +327,10 @@ class ExpressionScalar(Expression):
         return str(self._sympified_expression)
 
     def __repr__(self) -> str:
-        return 'Expression({})'.format(repr(self._original_expression))
+        if self._original_expression is None:
+            return f"ExpressionScalar('{self._sympified_expression!r}')"
+        else:
+            return f"ExpressionScalar({self._original_expression!r})"
 
     def __format__(self, format_spec):
         if format_spec == '':
@@ -274,20 +345,25 @@ class ExpressionScalar(Expression):
     def _sympify(cls, other: Union['ExpressionScalar', Number, sympy.Expr]) -> sympy.Expr:
         return other._sympified_expression if isinstance(other, cls) else sympify(other)
 
+    @classmethod
+    def _extract_sympified(cls, other: Union['ExpressionScalar', Number, sympy.Expr]) \
+                            -> Union['ExpressionScalar', Number, sympy.Expr]:
+        return getattr(other, '_sympified_expression', other)
+
     def __lt__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> Union[bool, None]:
-        result = self._sympified_expression < self._sympify(other)
+        result = self._sympified_expression < self._extract_sympified(other)
         return None if isinstance(result, sympy.Rel) else bool(result)
 
     def __gt__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> Union[bool, None]:
-        result = self._sympified_expression > self._sympify(other)
+        result = self._sympified_expression > self._extract_sympified(other)
         return None if isinstance(result, sympy.Rel) else bool(result)
 
     def __ge__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> Union[bool, None]:
-        result = self._sympified_expression >= self._sympify(other)
+        result = self._sympified_expression >= self._extract_sympified(other)
         return None if isinstance(result, sympy.Rel) else bool(result)
 
     def __le__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> Union[bool, None]:
-        result = self._sympified_expression <= self._sympify(other)
+        result = self._sympified_expression <= self._extract_sympified(other)
         return None if isinstance(result, sympy.Rel) else bool(result)
 
     def __eq__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> bool:
@@ -300,28 +376,28 @@ class ExpressionScalar(Expression):
         return hash(self._sympified_expression)
 
     def __add__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__add__(self._sympify(other)))
+        return self.make(self._sympified_expression.__add__(self._extract_sympified(other)))
 
     def __radd__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
         return self.make(self._sympify(other).__radd__(self._sympified_expression))
 
     def __sub__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__sub__(self._sympify(other)))
+        return self.make(self._sympified_expression.__sub__(self._extract_sympified(other)))
 
     def __rsub__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__rsub__(self._sympify(other)))
+        return self.make(self._sympified_expression.__rsub__(self._extract_sympified(other)))
 
     def __mul__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__mul__(self._sympify(other)))
+        return self.make(self._sympified_expression.__mul__(self._extract_sympified(other)))
 
     def __rmul__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__rmul__(self._sympify(other)))
+        return self.make(self._sympified_expression.__rmul__(self._extract_sympified(other)))
 
     def __truediv__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__truediv__(self._sympify(other)))
+        return self.make(self._sympified_expression.__truediv__(self._extract_sympified(other)))
 
     def __rtruediv__(self, other: Union['ExpressionScalar', Number, sympy.Expr]) -> 'ExpressionScalar':
-        return self.make(self._sympified_expression.__rtruediv__(self._sympify(other)))
+        return self.make(self._sympified_expression.__rtruediv__(self._extract_sympified(other)))
 
     def __neg__(self) -> 'ExpressionScalar':
         return self.make(self._sympified_expression.__neg__())
@@ -329,9 +405,15 @@ class ExpressionScalar(Expression):
     def __pos__(self):
         return self.make(self._sympified_expression.__pos__())
 
+    def _sympy_(self):
+        return self._sympified_expression
+
     @property
     def original_expression(self) -> Union[str, Number]:
-        return self._original_expression
+        if self._original_expression is None:
+            return str(self._sympified_expression)
+        else:
+            return self._original_expression
 
     @property
     def sympified_expression(self) -> sympy.Expr:
@@ -344,9 +426,35 @@ class ExpressionScalar(Expression):
         else:
             return serialized
 
+    def __getstate__(self):
+        return self.get_serialization_data()
+
+    def __setstate__(self, state):
+        self.__init__(state)
+
     def is_nan(self) -> bool:
         return sympy.sympify('nan') == self._sympified_expression
 
+    def evaluate_with_exact_rationals(self, scope: Mapping) -> Union[Number, numpy.ndarray]:
+        parsed_kwargs = self._parse_evaluate_numeric_arguments(scope)
+        result, self._exact_rational_lambdified = evaluate_lamdified_exact_rational(self.sympified_expression,
+                                                                                    self.variables,
+                                                                                    parsed_kwargs,
+                                                                                    self._exact_rational_lambdified)
+        try:
+            return _parse_evaluate_numeric(result)
+        except ValueError as err:
+            raise NonNumericEvaluation(self, result, scope) from err
+
+    def evaluate_in_scope(self, scope: Mapping) -> Union[Number, numpy.ndarray]:
+        parsed_kwargs = self._parse_evaluate_numeric_arguments(scope)
+        result, self._expression_lambda = evaluate_lambdified(self.underlying_expression, self.variables,
+                                                              parsed_kwargs, lambdified=self._expression_lambda)
+        try:
+            return _parse_evaluate_numeric(result)
+        except ValueError as err:
+            raise NonNumericEvaluation(self, result, scope) from err
+
 
 class ExpressionVariableMissingException(Exception):
     """An exception indicating that a variable value was not provided during expression evaluation.
@@ -372,7 +480,7 @@ class NonNumericEvaluation(Exception):
         qupulse.expressions.Expression.evaluate_numeric
     """
 
-    def __init__(self, expression: Expression, non_numeric_result: Any, call_arguments: Dict):
+    def __init__(self, expression: Expression, non_numeric_result: Any, call_arguments: Mapping):
         self.expression = expression
         self.non_numeric_result = non_numeric_result
         self.call_arguments = call_arguments
