@@ -7,19 +7,14 @@ import functools
 import warnings
 import logging
 
-try:
-    import tek_awg
-except ImportError:  # pragma: no cover
-    warnings.warn("Could not import Tektronix driver backend. "
-                  "If you wish to use it execute qupulse.hardware.awgs.install_requirements('tektronix')")
-    raise
+import tek_awg
 
 from qupulse.hardware.awgs.base import AWG, AWGAmplitudeOffsetHandling, ProgramOverwriteException
 from qupulse import ChannelID
 from qupulse._program._loop import Loop, make_compatible
 from qupulse._program.waveforms import Waveform as QuPulseWaveform
 from qupulse.utils.types import TimeType
-from qupulse.hardware.util import voltage_to_uint16, get_sample_times
+from qupulse.hardware.util import voltage_to_uint16, get_sample_times, traced
 from qupulse.utils import pairwise
 
 
@@ -238,6 +233,7 @@ class TektronixProgram:
         return self._amplitudes
 
 
+@traced
 class TektronixAWG(AWG):
     """Driver for Tektronix AWG object (5000/7000 series).
 
@@ -285,9 +281,6 @@ class TektronixAWG(AWG):
         super().__init__(identifier=identifier)
         self.logger = logger or logging.getLogger("qupulse.tektronix")
 
-        if device is None:
-            raise RuntimeError('Please install the tek_awg package or run "install_requirements" from this module')
-
         self._device = device
         self._synchronized = False # this gets set to True by synchronize or clear and to False on error during manupulation
 
@@ -609,7 +602,7 @@ class TektronixAWG(AWG):
         if previous_errors:
             self.logger.warning("Error queue not empty before sequence upload: %r", previous_errors)
 
-        positions_with_next = pairwise(positions, fillvalue=last_jump_to)
+        positions_with_next = pairwise(positions, zip_function=itertools.zip_longest, fillvalue=last_jump_to)
 
         self._synchronized = False
         for idx, ((element_index, next_element), sequencing_element) in enumerate(zip(positions_with_next, sequencing_elements)):
