@@ -7,7 +7,7 @@ from qupulse.utils.types import ChannelID, FrozenDict, FrozenMapping
 from qupulse.expressions import Expression, ExpressionScalar
 from qupulse.parameter_scope import Scope, MappedScope
 from qupulse.pulses.pulse_template import PulseTemplate, MappingTuple
-from qupulse.pulses.parameters import Parameter, MappedParameter, ParameterNotProvidedException, ParameterConstrainer
+from qupulse.pulses.parameters import ParameterNotProvidedException, ParameterConstrainer
 from qupulse._program.waveforms import Waveform
 from qupulse._program._loop import Loop
 from qupulse.serialization import Serializer, PulseRegistryType
@@ -112,7 +112,7 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
                                for k, v in template.channel_mapping.items()}
             template = template.template
 
-        self.__template = template
+        self.__template: PulseTemplate = template
         self.__parameter_mapping = FrozenDict(parameter_mapping)
         self.__external_parameters = set(itertools.chain(*(expr.variables for expr in self.__parameter_mapping.values())))
         self.__external_parameters |= self.constrained_parameters
@@ -201,7 +201,10 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
 
     @property
     def duration(self) -> Expression:
-        return self.__template.duration.evaluate_symbolic(self.__parameter_mapping)
+        return self.__template.duration.evaluate_symbolic(
+            {parameter_name: expression.underlying_expression
+             for parameter_name, expression in self.__parameter_mapping.items()}
+        )
 
     def get_serialization_data(self, serializer: Optional[Serializer]=None) -> Dict[str, Any]:
         data = super().get_serialization_data(serializer)
@@ -227,6 +230,9 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
 
         return data
 
+    def _is_atomic(self):
+        return self.__template._is_atomic()
+
     @classmethod
     def deserialize(cls,
                     serializer: Optional[Serializer]=None, # compatibility to old serialization routines, deprecated
@@ -237,7 +243,7 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
         # return MappingPulseTemplate(template=serializer.deserialize(template),
         #                             **kwargs)
 
-    def _validate_parameters(self, parameters: Dict[str, Union[Parameter, numbers.Real]], volatile: Set[str]):
+    def _validate_parameters(self, parameters: Dict[str, Union[numbers.Real]], volatile: Set[str]):
         missing = set(self.__external_parameters) - set(parameters.keys())
         if missing:
             raise ParameterNotProvidedException(missing.pop())
@@ -257,27 +263,11 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
         return {parameter: mapping_function.evaluate_numeric(**parameters)
                 for parameter, mapping_function in self.__parameter_mapping.items()}
 
-    def map_parameter_objects(self, parameters: Dict[str, Parameter],
-                              volatile: Set[str] = frozenset()) -> Dict[str, Parameter]:
-        """Map parameter objects (instances of Parameter class) according to the defined mappings.
-
-        Args:
-            parameters: Dictionary with parameter objects
-            volatile(Optional): Forwarded to `validate_parameter_constraints`
-        Returns:
-            A new dictionary with mapped parameter objects
-        """
-        self._validate_parameters(parameters=parameters, volatile=volatile)
-        return {parameter: MappedParameter(mapping_function,
-                                           {name: parameters[name] for name in mapping_function.variables})
-                for (parameter, mapping_function) in self.__parameter_mapping.items()}
-
     def map_scope(self, scope: Scope) -> MappedScope:
         return MappedScope(scope=scope, mapping=self.__parameter_mapping)
 
     def map_parameters(self,
-                       parameters: Dict[str, Union[Parameter, numbers.Real]]) -> Dict[str,
-                                                                                      Union[Parameter, numbers.Real]]:
+                       parameters: Dict[str, numbers.Real]) -> Dict[str, numbers.Real]:
         """Map parameter values according to the defined mappings.
 
         Args:
@@ -292,9 +282,6 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
         elif all(isinstance(parameter, numbers.Real) for parameter in parameters.values()):
             return self.map_parameter_values(parameters=parameters)
 
-        elif all(isinstance(parameter, Parameter) for parameter in parameters.values()):
-            return self.map_parameter_objects(parameters=parameters)
-
         else:
             raise TypeError('Values of parameter dict are neither all Parameter nor Real')
 
@@ -341,27 +328,30 @@ class MappingPulseTemplate(PulseTemplate, ParameterConstrainer):
             measurement_mapping=self.get_updated_measurement_mapping(measurement_mapping=measurement_mapping)
         )
 
-    @property
-    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
-        internal_integral = self.__template.integral
-        expressions = dict()
-
-        # sympy.subs() does not work if one of the mappings in the provided dict is an Expression object
-        # the following is an ugly workaround
-        # todo: make Expressions compatible with sympy.subs()
+    def _apply_mapping_to_inner_channel_dict(self, to_map: Dict[ChannelID, ExpressionScalar]) -> Dict[ChannelID, ExpressionScalar]:
         parameter_mapping = {parameter_name: expression.underlying_expression
                              for parameter_name, expression in self.__parameter_mapping.items()}
+        return {
+            self.__channel_mapping.get(ch, ch): ExpressionScalar(ch_expr.sympified_expression.subs(parameter_mapping,
+                                                                                                   simultaneous=True))
+            for ch, ch_expr in to_map.items()
+            if self.__channel_mapping.get(ch, ch) is not None
+        }
 
-        for channel, ch_integral in internal_integral.items():
-            channel_out = self.__channel_mapping.get(channel, channel)
-            if channel_out is None:
-                continue
+    @property
+    def integral(self) -> Dict[ChannelID, ExpressionScalar]:
+        return self._apply_mapping_to_inner_channel_dict(self.__template.integral)
 
-            expressions[channel_out] = ExpressionScalar(
-                ch_integral.sympified_expression.subs(parameter_mapping)
-            )
+    def _as_expression(self) -> Dict[ChannelID, ExpressionScalar]:
+        return self._apply_mapping_to_inner_channel_dict(self.__template._as_expression())
 
-        return expressions
+    @property
+    def initial_values(self) -> Dict[ChannelID, ExpressionScalar]:
+        return self._apply_mapping_to_inner_channel_dict(self.__template.initial_values)
+
+    @property
+    def final_values(self) -> Dict[ChannelID, ExpressionScalar]:
+        return self._apply_mapping_to_inner_channel_dict(self.__template.final_values)
 
 
 class MissingMappingException(Exception):
